@@ -1,4 +1,4 @@
-# Copyright Vahid Zehtab (vahid@zehtab.me) 2021
+# Copyright Vahid Zehtab (vahid@zehtab.me) 2022-2023
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -16,6 +16,7 @@ import typing as th
 from torch.utils.data import Dataset, DataLoader, random_split
 import lightning
 import dypy as dy
+from .transforms import transform_dataset, TransformsDescriptor, initialize_transforms
 
 
 class DataModule(lightning.LightningDataModule):
@@ -31,8 +32,13 @@ class DataModule(lightning.LightningDataModule):
         val_dataset_args: th.Optional[dict] = None,
         test_dataset: th.Optional[th.Union[str, Dataset]] = None,
         test_dataset_args: th.Optional[dict] = None,
+        # transforms
+        transforms: th.Optional[TransformsDescriptor] = None,
+        train_transforms: th.Optional[TransformsDescriptor] = None,
+        val_transforms: th.Optional[TransformsDescriptor] = None,
+        test_transforms: th.Optional[TransformsDescriptor] = None,
         # batch_size
-        batch_size: th.Optional[int] = 16,
+        batch_size: th.Optional[int] = None,
         train_batch_size: th.Optional[int] = None,
         val_batch_size: th.Optional[int] = None,
         test_batch_size: th.Optional[int] = None,
@@ -42,14 +48,19 @@ class DataModule(lightning.LightningDataModule):
         val_pin_memory: th.Optional[bool] = None,
         test_pin_memory: th.Optional[bool] = None,
         # shuffle
-        train_shuffle: bool = True,
-        val_shuffle: bool = False,
-        test_shuffle: bool = False,
+        train_shuffle: th.Optional[bool] = None,
+        val_shuffle: th.Optional[bool] = None,
+        test_shuffle: th.Optional[bool] = None,
         # num_workers
         num_workers: int = 0,
         train_num_workers: th.Optional[int] = None,
         val_num_workers: th.Optional[int] = None,
         test_num_workers: th.Optional[int] = None,
+        # generic dataloader params
+        dataloader_args: th.Optional[dict] = None,
+        train_dataloader_args: th.Optional[dict] = None,
+        val_dataloader_args: th.Optional[dict] = None,
+        test_dataloader_args: th.Optional[dict] = None,
         # seed
         seed: int = 0,
         # initialization settings
@@ -93,13 +104,22 @@ class DataModule(lightning.LightningDataModule):
             **(test_dataset_args or dict()),
         }
 
+        # transforms
+        transforms = initialize_transforms(transforms, force_list=True)
+        self.train_transforms = (
+            initialize_transforms(train_transforms, force_list=True) if train_transforms is not None else transforms
+        )
+        self.val_transforms = (
+            initialize_transforms(val_transforms, force_list=True) if val_transforms is not None else transforms
+        )
+        self.test_transforms = (
+            initialize_transforms(test_transforms, force_list=True) if test_transforms is not None else transforms
+        )
+
         # batch_size
         self.train_batch_size = train_batch_size if train_batch_size is not None else batch_size
         self.val_batch_size = val_batch_size if val_batch_size is not None else batch_size
         self.test_batch_size = test_batch_size if test_batch_size is not None else batch_size
-        assert (
-            self.train_batch_size or self.val_batch_size or self.test_batch_size
-        ), "at least one of batch_sizes should be a positive number"
 
         # pin memory
         self.train_pin_memory = train_pin_memory if train_pin_memory is not None else pin_memory
@@ -114,6 +134,12 @@ class DataModule(lightning.LightningDataModule):
         self.val_num_workers = val_num_workers if val_num_workers is not None else num_workers
         self.test_num_workers = test_num_workers if test_num_workers is not None else num_workers
 
+        # generic dataloader params
+        base_dataloader_args = dataloader_args or dict()
+        self.train_dataloader_args = {**base_dataloader_args, **(train_dataloader_args or dict())}
+        self.val_dataloader_args = {**base_dataloader_args, **(val_dataloader_args or dict())}
+        self.test_dataloader_args = {**base_dataloader_args, **(test_dataloader_args or dict())}
+
     @staticmethod
     def get_dataset(dataset: th.Union[str, Dataset], **params):
         assert isinstance(
@@ -126,16 +152,24 @@ class DataModule(lightning.LightningDataModule):
         else:
             return DataModule.get_dataset(dy.get_value(dataset, strict=True), **params)
 
-    def setup(self, stage: th.Optional[str] = None) -> None:
+    def setup(self, stage: th.Optional[str] = None, transform: bool = True) -> None:
         if stage == "fit" or stage == "tune":
             self.split_dataset()
+            if transform:
+                self.train_data = transform_dataset(self.train_data, self.train_transforms)
+                self.val_data = transform_dataset(self.val_data, self.val_transforms)
 
-        elif stage == "test" and self.test_batch_size:
+            if self.val_data is None:
+                self.val_dataloader = None
+
+        elif stage == "test" and self.test_batch_size and self.test_data is not None:
             test_data = self.get_dataset(
                 self.test_dataset,
                 **self.test_dataset_args,
             )
             self.test_data = test_data
+            if transform:
+                self.test_data = transform_dataset(self.test_data, self.test_transforms)
 
     def split_dataset(self):
         self.train_data = self.train_data or self.dataset  # in case the train_data is not provided
@@ -144,7 +178,7 @@ class DataModule(lightning.LightningDataModule):
             self.train_dataset,
             **self.train_dataset_args,
         )
-        self.data = self.train_data  # for ease of access (e.g. for plotting)
+        self.data = dataset  # for ease of access (e.g. for plotting)
 
         if not self.val_size:
             # setup train data (in case validation is not to be a subset of provided dataset with val_size)
@@ -173,6 +207,11 @@ class DataModule(lightning.LightningDataModule):
         shuffle = shuffle if shuffle is not None else getattr(self, f"{name}_shuffle")
         num_workers = num_workers if num_workers is not None else getattr(self, f"{name}_num_workers")
         pin_memory = pin_memory if pin_memory is not None else getattr(self, f"{name}_pin_memory")
+        params = {**getattr(self, f"{name}_dataloader_args"), **params}
+
+        # set fractional batches if batch_size is negative
+        batch_size = (len(data) // abs(batch_size)) if batch_size is not None and batch_size < 0 else batch_size
+
         return DataLoader(
             dataset=data,
             batch_size=batch_size,
