@@ -1,4 +1,4 @@
-# Copyright Vahid Zehtab (vahid@zehtab.me) 2021
+# Copyright Vahid Zehtab (vahid@zehtab.me) 2022-2023
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -21,6 +21,8 @@ from lightning_toolbox.objective_function import Objective
 import torch
 from .datastructures import ArgsListDict
 
+DICT_TYPE = th.Union[dict, th.Any]
+
 
 class TrainingModule(lightning.LightningModule):
     def __init__(
@@ -28,11 +30,11 @@ class TrainingModule(lightning.LightningModule):
         # model
         model: th.Optional[torch.nn.Module] = None,  # model instance
         model_cls: th.Optional[str] = None,
-        model_args: th.Optional[dict] = None,
+        model_args: th.Optional[DICT_TYPE] = None,
         # objective
         objective: th.Optional[Objective] = None,  # objective instance
-        objective_cls: th.Union[type, str] = "lightning_toolbox.Objective",
-        objective_args: th.Optional[dict] = None,
+        objective_cls: str = "lightning_toolbox.Objective",  # v0.0.29: remove type annotation to avoid CLI auto-discovery
+        objective_args: th.Optional[DICT_TYPE] = None,
         # optimization configs
         # optimizer name or class
         optimizer: th.Union[str, type, th.List[th.Union[str, type]], None] = None,
@@ -40,7 +42,7 @@ class TrainingModule(lightning.LightningModule):
         optimizer_is_active: th.Optional[th.Union[dy.FunctionDescriptor, th.List[dy.FunctionDescriptor]]] = None,
         # optimizer parameters (self.<*>)
         optimizer_parameters: th.Optional[th.Union[th.List[str], str]] = None,
-        optimizer_args: th.Optional[dict] = None,
+        optimizer_args: th.Union[DICT_TYPE, th.List[th.Optional[DICT_TYPE]], None] = None,
         # learning rate
         lr: th.Union[th.List[float], float] = 1e-4,
         # schedulers
@@ -49,7 +51,7 @@ class TrainingModule(lightning.LightningModule):
         scheduler_name: th.Optional[th.Union[str, th.List[str]]] = None,
         # optimizer index
         scheduler_optimizer: th.Optional[th.Union[int, th.List[int]]] = None,
-        scheduler_args: th.Optional[th.Union[dict, th.List[dict]]] = None,
+        scheduler_args: th.Optional[th.Union[DICT_TYPE, th.List[DICT_TYPE]]] = None,
         scheduler_interval: th.Union[str, th.List[str]] = "epoch",
         scheduler_frequency: th.Union[int, th.List[int]] = 1,
         scheduler_monitor: th.Optional[th.Union[str, th.List[str]]] = None,
@@ -57,6 +59,7 @@ class TrainingModule(lightning.LightningModule):
         # initialization settings
         save_hparams: bool = True,
         initialize_superclass: bool = True,
+        pretrained_state: th.Optional[str] = None,
     ) -> None:
         if initialize_superclass:
             super().__init__()
@@ -125,6 +128,12 @@ class TrainingModule(lightning.LightningModule):
         if model is not None or model_args is not None or model_cls is not None:
             self.model = model if model is not None else dy.eval(model_cls)(**(model_args or dict()))
 
+        # load pretrained checkpoint
+        # TODO: add smart loading of state dict, e.g. if the checkpoint has different structure,
+        # load only the common parts and ignore the rest
+        if pretrained_state is not None:
+            self.load_state_dict(torch.load(pretrained_state)["state_dict"])
+
     @functools.cached_property
     def __optimizers_is_active_list(self):
         if self.optimizers_list is None:
@@ -195,7 +204,7 @@ class TrainingModule(lightning.LightningModule):
             return None  # this will skip this step, but the optimizer position will be incremented
             # TODO: check if how this affects the schedulers
 
-        results, factors = self.objective(
+        _return_val = self.objective(
             batch=batch,
             optimizer_idx=optimizer_idx,
             batch_idx=batch_idx,
@@ -203,11 +212,14 @@ class TrainingModule(lightning.LightningModule):
             return_factors=True,
             **kwargs,
         )
+        results, factors = _return_val if isinstance(_return_val, tuple) else (_return_val, None)
         if log_results:
             self.log_step_results(results, factors, name)
         if return_results:
             return (results, factors) if return_factors else results
-        return results["loss"] if not is_val else None
+        if not is_val:
+            loss = results["loss"]
+            return loss.mean() if isinstance(loss, torch.Tensor) else loss
 
     def configure_optimizers(self):
         """Initialize the optimizers and learning rate schedulers."""
@@ -301,8 +313,8 @@ class TrainingModule(lightning.LightningModule):
                 sync_dist=True,
             )
 
-    def training_step(self, batch, batch_idx, optimizer_idx=None, **kwargs):
-        return self.step(batch, batch_idx, optimizer_idx, name="train", **kwargs)
+    def training_step(self, batch, batch_idx, **kwargs):
+        return self.step(batch, batch_idx, name="train", **kwargs)
 
     def validation_step(self, batch, batch_idx, **kwargs):
         return self.step(batch, batch_idx, name="val", **kwargs)
